using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using symdump.symfile.util;

namespace symdump.symfile
{
    public class TypeInfo : IEquatable<TypeInfo>
    {
        public readonly ClassType ClassType;
        public readonly uint[] Dims;
        public readonly uint Size;
        public string Tag;
        public readonly TypeDef TypeDef;
        public readonly int SYMFileSequenceNumber;
        public string SLDSourceFileName;
        public string ObjectFileName;
        public bool OriginalWasFake;

        public TypeInfo(BinaryReader reader, bool extended, BenLincoln.DebugSymbols.SequenceIssuer issuer, SymbolParsingTracker tracker, 
            bool renameForCompatibility, bool autoRenameFakes)
        {
            // comments in this section refer to the decompiled version of DUMPSYM.EXE
            // https://github.com/lab313ru/dumpsym_src/blob/master/main.c
            ClassType = reader.ReadClassType();             // sub_39B(read_word(input));
            TypeDef = reader.ReadTypeDef(issuer);           // sub_504(read_word(input));
            Console.WriteLine(string.Format("Debug: file offset = {0:X8}.", reader.BaseStream.Position));
            Console.WriteLine(string.Format("Debug: size = {0}.", Size));
            Size = reader.ReadUInt32();                     // printf("size %ld ", read_dword(input));
            SYMFileSequenceNumber = issuer.GetNextNumber();

            if (extended)
            {
                Console.WriteLine(string.Format("Debug: file offset = {0:X8}.", reader.BaseStream.Position));
                var n = reader.ReadUInt16();
                Console.WriteLine(string.Format("Debug: n = {0}.", n));
                Dims = new uint[n];
                for (var i = 0; i < n; ++i)
                {
                    Console.WriteLine(string.Format("Debug: file offset = {0:X8}.", reader.BaseStream.Position));
                    Dims[i] = reader.ReadUInt32();
                    Console.WriteLine(string.Format("Debug: Dims[{0}] = {1:X8}.", i, Dims[i]));
                }

                Console.WriteLine(string.Format("Debug: file offset = {0:X8}.", reader.BaseStream.Position));
                Tag = reader.ReadPascalString();
                Console.WriteLine(string.Format("Debug: got tag '{0}'.", Tag));
                if (Tag == ".0fake")
                {
                    Console.WriteLine("Debug");
                }
                OriginalWasFake = IsFake;
                if (renameForCompatibility)
                {
                    if (OriginalWasFake)
                    {
                        Tag = BenLincoln.DebugSymbols.SymDump.CElement.GetCompatibleName(Tag, true);
                    }
                }
                if (autoRenameFakes)
                {
                    if (OriginalWasFake)
                    {
                        string lastObjectFileName = "";
                        if (tracker.ObjectFileDataEntries.Count > 0)
                        {
                            lastObjectFileName = tracker.ObjectFileDataEntries[tracker.ObjectFileDataEntries.Count - 1].FileName;
                        }
                        string lastSLDSourceFileName = "";
                        if (tracker.SLDEntries.Count > 0)
                        {
                            lastSLDSourceFileName = tracker.SLDEntries[tracker.SLDEntries.Count - 1].SLDFile;
                        }
                        Tag = BenLincoln.DebugSymbols.Utilities.GetNewFakeName(Tag, tracker.CurrentObjectFileData.FileName, tracker.CurrentSLDFile, lastObjectFileName, lastSLDSourceFileName);
                    }
                }
                //bool doneGettingTags = false;
                //if (Tag == "")
                //{
                //    doneGettingTags = true;
                //}
                //while (!doneGettingTags)
                //{
                //    string nextTag = reader.ReadPascalString();
                //    if (nextTag == "")
                //    {
                //        doneGettingTags = true;
                //    }
                //    else
                //    {
                //        Console.WriteLine(string.Format("Debug: got additional tag '{0}'.", nextTag));
                //    }
                //}
            }
            else
            {
                Dims = new uint[0];
                Tag = null;
            }
        }

        public bool IsFake => Tag != null && new Regex(@"^(\.|_)\d+fake$").IsMatch(Tag);

        public bool Equals(TypeInfo other)
        {
            if (ReferenceEquals(null, other)) return false;
            if (ReferenceEquals(this, other)) return true;

            if ((OriginalWasFake) && (other.OriginalWasFake))
            {
                return ClassType == other.ClassType && TypeDef.Equals(other.TypeDef) && Size == other.Size &&
                       Dims.SequenceEqual(other.Dims) && other.OriginalWasFake;
            }
            return ClassType == other.ClassType && TypeDef.Equals(other.TypeDef) && Size == other.Size &&
                   Dims.SequenceEqual(other.Dims) && string.Equals(Tag, other.Tag);
        }

        public override string ToString()
        {
            return $"classType={ClassType} typeDef={TypeDef} size={Size}, dims=[{string.Join(",", Dims)}] // {ObjectFileName}";
        }

        public string AsCode(string name)
        {
            return TypeDef.AsCode(name, this).Trim();
        }

        public string GetCType()
        {
            return TypeDef.GetCType(this, ToString());
        }

        public string GetName(string name)
        {
            return TypeDef.GetName(name, this);
        }

        public List<string> GetDerivedTypes(string name)
        {
            return TypeDef.GetDerivedTypeList(name, this);
        }

        public BenLincoln.DebugSymbols.SymDump.CTypeDef ToBLDSCTypeDef(string name, string defaultSourceFile)
        {
            BenLincoln.DebugSymbols.SymDump.CTypeDef result = new BenLincoln.DebugSymbols.SymDump.CTypeDef();

            result.name = GetName(name);
            result.c_type = GetCType();
            result.sym_file_sequence_number = SYMFileSequenceNumber;
            result.source_file = defaultSourceFile;
            result.fake = OriginalWasFake;
            if ((SLDSourceFileName != null) && (SLDSourceFileName != ""))
            {
                Console.WriteLine(string.Format("Info: using SLD source file '{0}' for typedef '{2}'", SLDSourceFileName, defaultSourceFile, result.name));
                result.source_file = SLDSourceFileName;
            }
            result.object_file = ObjectFileName;
            //result.hashcode = GetHashCode();
            result.typedef_signature = AsCode(name);

            return result;
        }

        public string DumpJSON(string name, string sourceFile, string lineIndent)
        {
            BenLincoln.DebugSymbols.SymDump.CTypeDef resultDef = ToBLDSCTypeDef(name, sourceFile);

            return BenLincoln.DebugSymbols.Utilities.IndentString(resultDef.ToJSON(), lineIndent);
            //StringBuilder builder = new StringBuilder();

            //builder.Append(lineIndent);
            //builder.Append("{");
            //builder.Append(Environment.NewLine);
            //builder.Append(lineIndent);
            //builder.Append(String.Format("\t\"name\": \"{0}\",", name));
            //builder.Append(Environment.NewLine);
            //if (sourceFile != "")
            //{
            //    builder.Append(lineIndent);
            //    builder.Append(String.Format("\t\"source_file\": \"{0}\",", sourceFile));
            //    builder.Append(Environment.NewLine);
            //}
            //builder.Append(lineIndent);
            //builder.Append(String.Format("\t\"definition\": \"{0}\",", AsCode(name)));
            //builder.Append(Environment.NewLine);
            //builder.Append(lineIndent);
            //builder.Append("}");

            //return builder.ToString();
        }

        public override bool Equals(object obj)
        {
            if (ReferenceEquals(null, obj)) return false;
            if (ReferenceEquals(this, obj)) return true;
            if (obj.GetType() != GetType()) return false;
            return Equals((TypeInfo) obj);
        }

        public override int GetHashCode()
        {
            unchecked
            {
                var hashCode = (int) ClassType;
                hashCode = (hashCode * 397) ^ TypeDef.GetHashCode();
                hashCode = (hashCode * 397) ^ (int) Size;
                hashCode = (hashCode * 397) ^ Dims.GetHashCode();
                if (Tag != null)
                {
                    hashCode = (hashCode * 397) ^ (!IsFake ? Tag.GetHashCode() : 0);
                }
                return hashCode;
            }
        }
    }
}
